import logging
from ipaddress import IPv4Address
from typing import Union

from pythonping import ping

logger = logging.getLogger(__name__)


def wait_ping(host: Union[str, IPv4Address], timeout: int = 3, count: int = 5, verbose: bool = False) -> bool:
    """
    Wait until the remote host is pingable.

    :param host: Remote host
    :param timeout: Timeout in seconds.
    :param count: Number of pings. If -1, ping endlessly.
    :param verbose: Verbose output.
    :return: True if the remote host is pingable, False otherwise.
    """

    if type(host) is IPv4Address:
        host = str(host)

    logger.debug(f"Wait until '{host}' is pingable.")

    # ping remote host
    if count == -1:
        _count = 1
        while True:
            ping_result = ping(host, timeout=timeout, count=_count, verbose=verbose)
            if ping_result.success():
                break
    else:
        _count = count
        ping_result = ping(host, timeout=timeout, count=_count, verbose=verbose)

    if not ping_result.success():
        logger.error(f"Remote host '{host}' is not pingable.")
        return False

    logger.debug(f"Remote host '{host}' is pingable.")
    return True



import json
from pathlib import Path
from typing import Any, Union

from pydantic import BaseModel, create_model


class FileConfig(BaseModel):
    def __init__(self,
                 file_path: Union[Path, str, None] = None,
                 file_postfix: str = ".json",
                 file_must_exist: bool = False,
                 **overwrite_data: Any):
        if file_path is None:
            file_path = Path(Path.cwd() / self.__class__.__name__.lower()).with_suffix(file_postfix)
        else:
            file_path = Path(file_path)
            if file_path.suffix == "":
                file_path = file_path.with_suffix(file_postfix)

        # read data from file
        if file_path.is_file():
            with file_path.open("r", encoding="utf-8") as file:
                data = json.load(file)
        elif file_path.is_dir():
            raise ValueError(f"{self.__class__.__name__} file path '{file_path}' is a directory.")
        else:
            if file_must_exist:
                raise FileNotFoundError(f"{self.__class__.__name__} file '{file_path}' not found.")
            data = {}

        # overwrite data
        for k, v in overwrite_data.items():
            data[k] = v

        super().__init__(**data)

        self._file_path = file_path

    @property
    def file_path(self) -> Path:
        """
        File path

        :return: Path
        """

        return self._file_path

    def save(self, validate: bool = True, indent: int = 4, encoding: str = "utf-8"):
        if validate:
            validate_model_info = {}
            for field_name, field_info in self.model_fields.items():
                validate_model_info[field_name] = (field_info.annotation, field_info)
            validate_model = create_model(f"{self.__class__.__name__}_Validate", **validate_model_info)

            params = self.model_dump()
            validated = validate_model(**params)
            self_json = validated.model_dump_json(indent=indent)
        else:
            self_json = self.model_dump_json(indent=indent)

        with self.file_path.open("w", encoding=encoding) as file:
            file.write(self_json)



import logging
import os
import shutil
from pathlib import Path
from typing import Union

import bson
from pymongo.database import Database

logger = logging.getLogger(__name__)


def dump(db: Database, path: Union[str, Path], overwrite: bool = False) -> None:
    """
    MongoDB Dump
    :param db: MongoDB database
    :param path: Database dump path
    :param overwrite: Overwrite existing files
    :return: None
    """

    # convert path to Path object
    if type(path) is str:
        path = Path(path)

    logger.debug(f"Dumping database to '{path}'")

    if path.exists():
        if not overwrite:
            raise FileExistsError(f"Path '{path}' already exists")
        if path.is_file():
            # remove existing file
            logger.debug(f"Removing existing file '{path}'")
            os.remove(path)
        else:
            # remove existing files
            logger.debug(f"Removing existing files in '{path}'")
            shutil.rmtree(path)

    # ensure path dir exists
    path.with_suffix("").mkdir(parents=True, exist_ok=True)

    for collection_name in db.list_collection_names():
        logger.debug(f"Dumping collection '{collection_name}'")

        # get all documents from collection
        documents_encoded = []
        for document in db[collection_name].find():
            # encode document to bson
            document_encode = bson.BSON.encode(document)
            documents_encoded.append(document_encode)

        logger.debug(f"Dumping {len(documents_encoded)} documents")

        # create bson file
        bson_file = path.with_suffix("") / f'{collection_name}.bson'
        with open(bson_file, 'wb+') as f:
            # write all documents to bson file
            for document_encode in documents_encoded:
                f.write(document_encode)

    if path.suffix == ".zip":
        # zip dump
        logger.debug(f"Zipping dump '{path.with_suffix('')}' as '{path}'")
        shutil.make_archive(str(path.with_suffix("")), path.suffix.replace(".", ""), path.with_suffix(""))
        logger.debug("Zipped dump successfully")

        # remove bson files
        logger.debug(f"Removing bson files in '{path.with_suffix('')}'")
        shutil.rmtree(path.with_suffix(""))

    logger.debug("Database dumped successfully")


def restore(db: Database, path: Union[str, Path], overwrite: bool = False) -> None:
    """
    MongoDB Restore
    :param db: Database connection
    :param path: Database dump path
    :param overwrite: Overwrite existing collections
    :return: None
    """

    # convert path to Path object
    if type(path) is str:
        path = Path(path)

    logger.debug(f"Restoring database from '{path}'")

    # unzip dump
    if path.suffix == ".zip":
        if path.with_suffix("").is_dir():
            logger.debug(f"Removing existing directory '{path.with_suffix('')}'")
            shutil.rmtree(path.with_suffix(""))
        logger.debug(f"Unzipping dump '{path}'")
        shutil.unpack_archive(path, path.with_suffix(""))
        logger.debug("Unzipped dump successfully")

    # check if path is a directory
    if not os.path.isdir(path.with_suffix("")):
        raise FileNotFoundError(f"Path '{path.with_suffix('')}' does not exist")

    for bson_file in os.listdir(path.with_suffix("")):
        bson_file = path.with_suffix("") / bson_file

        # check bson file is a file
        if not os.path.isfile(bson_file):
            continue

        # check bson file is a bson file
        if not bson_file.suffix == ".bson":
            continue

        # get collection name
        collection_name = bson_file.with_suffix("").name

        logger.debug(f"Restoring collection '{collection_name}'")

        with open(bson_file, 'rb+') as f:
            # read bson file
            bson_file_data = f.read()

            # decode bson file
            documents = bson.decode_all(bson_file_data)

        # drop collection if exists
        if collection_name in db.list_collection_names():
            if overwrite:
                logger.debug(f"Dropping collection '{collection_name}'")
                db[collection_name].drop()
            else:
                raise FileExistsError(f"Collection '{collection_name}' already exists")

        # insert documents into collection
        logger.debug(f"Restoring {len(documents)} documents")
        db[collection_name].insert_many(documents)

    # remove bson files
    if path.suffix == ".zip":
        logger.debug(f"Removing bson files in '{path.with_suffix('')}'")
        shutil.rmtree(path.with_suffix(""))

    logger.debug("Database restored successfully")



from mongoengine import EmbeddedDocument, StringField, IntField


class HashedPasswordDocument(EmbeddedDocument):
    encoding = StringField()
    hash_function = StringField()
    interactions = IntField()
    key_length = IntField()
    salt = StringField()
    hashed_password = StringField()



import ctypes
import logging
import os
from functools import wraps
from typing import Optional

logger = logging.getLogger(__name__)


def is_admin() -> bool:
    """
    Check if the current user has admin privileges.

    :return: True if the user has admin privileges, False otherwise
    """

    logger.debug("Check if the current user has admin privileges.")

    try:
        _is_admin = os.getuid() == 0
    except AttributeError:
        _is_admin = ctypes.windll.shell32.IsUserAnAdmin() != 0

    logger.debug(f"Current user {'has' if _is_admin else 'does not have'} admin privileges.")

    return _is_admin


def require_admin(show_window: Optional[int] = None):
    def decorator(func):
        """
        Decorator to require admin rights
        """

        @wraps(func)
        def wrapper(*args, **kwargs):
            """
            Wrapper function
            """

            # check if is admin
            if not is_admin():
                logger.error("This function requires admin rights.")
                raise SystemExit(1)

            return func(*args, **kwargs)

        return wrapper

    return decorator



from wiederverwendbar.starlette_admin.view import PropertyModelView
from wiederverwendbar.starlette_admin.view import property_field

from wiederverwendbar.starlette_admin.mongoengine import Admin

from wiederverwendbar.starlette_admin.mongoengine import FixedModelView

from wiederverwendbar.starlette_admin.mongoengine import GenericEmbeddedConverter
from wiederverwendbar.starlette_admin.mongoengine import GenericEmbeddedDocumentField

from wiederverwendbar.starlette_admin.mongoengine import IPv4Converter



from jinja2 import PackageLoader
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from starlette_admin.contrib.mongoengine import Admin as BaseAdmin


class Admin(BaseAdmin):
    def init_routes(self) -> None:
        super().init_routes()

        # find the statics mount index
        statics_index = None
        for i, route in enumerate(self.routes):
            if isinstance(route, Mount) and route.name == "statics":
                statics_index = i
                break
        if statics_index is None:
            raise ValueError("Could not find statics mount")

        # override the static files route
        self.routes[statics_index] = Mount("/statics",
                                           app=StaticFiles(
                                               directory=self.statics_dir,
                                               packages=[("wiederverwendbar", "starlette_admin/statics"), "starlette_admin"]),
                                           name="statics")

    def _setup_templates(self) -> None:
        super()._setup_templates()
        additional_loaders = [
            PackageLoader("wiederverwendbar", "starlette_admin/mongoengine/generic_embedded_document_field/templates"),
        ]

        self.templates.env.loader.loaders.extend(additional_loaders)



from typing import Callable, Union


def find_class_method(bases, method_name) -> Union[Callable, None]:
    if hasattr(bases, "__iter__"):
        for base in bases:
            if hasattr(base, method_name):
                return getattr(base, method_name)
            for _bases in base.__bases__:
                found = find_class_method(bases=_bases, method_name=method_name)
                if found is not None:
                    return found
    return None



from pydantic import BaseModel
from pydantic import Field as BaseField


# noinspection PyPep8Naming
def Field(*args, secret: bool = False, **kwargs) -> BaseField:
    """This method does the same as :py:class:`pydantic.Field`, but adds the possibility to hide the value."""

    if secret:
        kwargs.update(secret=True)

    return BaseField(*args, **kwargs)


class PrintableSettings(BaseModel):
    """
    Base class for settings that can be printed.
    Secrets are hidden.
    """

    def __repr_args__(self):
        for k, v in super().__repr_args__():
            field = self.model_fields.get(k)
            json_scheme_extra = field.json_schema_extra if field.json_schema_extra is not None else {}
            secret = json_scheme_extra.get("secret", False)
            if secret:
                yield k, "********"
            else:
                yield k, v



from ipaddress import IPv4Address

from mongoengine.base.fields import BaseField


class IPv4AddressField(BaseField):
    def validate(self, value, clean=True):
        super().validate(value, clean)
        try:
            IPv4Address(value)
        except ValueError:
            self.error(f"Invalid IPv4 address: {value}")

    def to_mongo(self, value):
        return str(value)

    def to_python(self, value):
        return IPv4Address(value)



from typing import Any, Union

from mongoengine import Document


class PropertyDocument(Document):
    meta = {"abstract": True}

    def __getitem__(self, item: Union[int, str]) -> Any:
        """
        Get item from model

        :param item: Index or key
        :return: Index value or key value
        """

        # check if is a property
        if hasattr(self, item):
            attr = getattr(self.__class__, item)
            if isinstance(attr, property):
                return getattr(self, item)

        return super().__getitem__(item)




from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field.converter import GenericEmbeddedConverter
from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field.field import GenericEmbeddedDocumentField



from typing import Any

import mongoengine as me
import starlette_admin as sa
from starlette_admin.contrib.mongoengine.converters import ModelConverter, converts
from starlette_admin.helpers import slugify_class_name

from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field.field import GenericEmbeddedDocumentField, ListField


class GenericEmbeddedConverter(ModelConverter):
    @converts(me.GenericEmbeddedDocumentField)
    def conv_generic_embedded_document_field(self, *args: Any, **kwargs: Any) -> sa.BaseField:
        common = self._field_common(*args, **kwargs)
        field = kwargs["field"]
        if field.choices is None:
            raise ValueError("GenericEmbeddedDocumentField requires embedded_docs")
        embedded_doc_name_mapping = {}
        embedded_doc_fields = {}
        for doc in field.choices:
            if not issubclass(doc, me.EmbeddedDocument):
                raise ValueError(f"{doc} is not a subclass of EmbeddedDocument")
            doc_meta = getattr(doc, "_meta", {})
            name = doc_meta.get("name", doc.__name__)
            embedded_doc_name_mapping[name] = doc
            doc_fields = []
            for _field in getattr(doc, "_fields_ordered"):
                if _field.startswith("_"):
                    continue
                kwargs["field"] = getattr(doc, _field)
                doc_fields.append(self.convert(*args, **kwargs))
            embedded_doc_fields[name] = doc_fields
        return GenericEmbeddedDocumentField(**common, embedded_doc_fields=embedded_doc_fields, embedded_doc_name_mapping=embedded_doc_name_mapping)

    @converts(me.ListField, me.SortedListField)
    def conv_list_field(self, *args: Any, **kwargs: Any) -> sa.BaseField:
        field = kwargs["field"]
        if field.field is None:
            raise ValueError(f'ListField "{field.name}" must have field specified')
        if isinstance(
                field.field,
                (me.ReferenceField, me.CachedReferenceField, me.LazyReferenceField),
        ):
            """To Many reference"""
            dtype = field.field.document_type_obj
            identity = slugify_class_name(
                dtype if isinstance(dtype, str) else dtype.__name__
            )
            return sa.HasMany(**self._field_common(*args, **kwargs), identity=identity)
        field.field.name = field.name
        kwargs["field"] = field.field
        if isinstance(field.field, (me.DictField, me.MapField)):
            return self.convert(*args, **kwargs)
        if isinstance(field.field, me.EnumField):
            admin_field = self.convert(*args, **kwargs)
            admin_field.multiple = True  # type: ignore [attr-defined]
            return admin_field
        return ListField(self.convert(*args, **kwargs), required=field.required)



from wiederverwendbar.starlette_admin.mongoengine.admin import Admin

from wiederverwendbar.starlette_admin.mongoengine.view import FixedModelView

from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field import GenericEmbeddedConverter
from wiederverwendbar.starlette_admin.mongoengine.generic_embedded_document_field import GenericEmbeddedDocumentField

from wiederverwendbar.starlette_admin.mongoengine.ipv4_field import IPv4Converter



from typing import Any

import starlette_admin as sa
from starlette_admin.contrib.mongoengine.converters import ModelConverter, converts

from wiederverwendbar.mongoengine.fields.ipv4_address_field import IPv4AddressField
from wiederverwendbar.mongoengine.fields.ipv4_network_field import IPv4NetworkField


class IPv4Converter(ModelConverter):
    @converts(IPv4AddressField, IPv4NetworkField)
    def conv_ipv4_field(self, *args: Any, **kwargs: Any) -> sa.BaseField:
        return sa.StringField(**self._field_common(*args, **kwargs))



import random


def animal_name_generator() -> list[str]:
    animals = [
        ["Dog", "Hund"], ["Puppy", "Welpe"], ["Turtle", "Schildkröte"], ["Parrot", "Papagei"], ["Cat", "Katze"], ["Kitten", "Kätzchen"],
        ["Goldfish", "Goldfisch"], ["Mouse", "Maus"], ["Hamster", "Hamster"], ["Cow", "Kuh"], ["Rabbit", "Kaninchen"], ["Ducks", "Enten"],
        ["Shrimp", "Garnele"], ["Pig", "Schwein"], ["Goat", "Ziege"], ["Crab", "Krabbe"], ["Deer", "Hirsche"], ["Bee", "Biene"], ["Sheep", "Schaf"],
        ["Fish", "Fisch"], ["Turkey", "Truthahn"], ["Dove", "Taube"], ["Chicken", "Huhn"], ["Horse", "Pferd"], ["Crow", "Krähe"], ["Peacock", "Pfau"],
        ["Dove", "Taube"], ["Sparrow", "Sperling"], ["Goose", "Gans"], ["Stork", "Storch"], ["Pigeon", "Taube"], ["Turkey", "Truthahn"], ["Hawk", "Falke"],
        ["BaldEagle", "Weißkopfseeadler"], ["Raven", "Rabe"], ["Parrot", "Papagei"], ["Flamingo", "Flamingo"], ["Seagull", "Möwe"], ["Ostrich", "Strauß"],
        ["Swallow", "Schwalbe"], ["BlackBird", "Schwarzvogel"], ["Penguin", "Pinguin"], ["Robin", "Rotkehlchen"], ["Swan", "Schwan"], ["Owl", "Eule"],
        ["Woodpecker", "Specht"], ["Squirrel", "Eichhörnchen"], ["Dog", "Hund"], ["Chimpanzee", "Schimpanse"], ["Ox", "Ochse"], ["Lion", "Löwe"],
        ["Panda", "Panda"], ["Walrus", "Walross"], ["Otter", "Otter"], ["Mouse", "Maus"], ["Kangaroo", "Känguru"], ["Goat", "Ziege"], ["Horse", "Pferd"],
        ["Monkey", "Affe"], ["Cow", "Kuh"], ["Koala", "Koala"], ["Mole", "Maulwurf"], ["Elephant", "Elefant"], ["Leopard", "Leopard"],
        ["Hippopotamus", "Nilpferd"], ["Giraffe", "Giraffe"], ["Fox", "Fuchs"], ["Coyote", "Kojote"], ["Hedgehong", "Igelhund"], ["Sheep", "Schaf"],
        ["Deer", "Hirsche"], ["Giraffe", "Giraffe"], ["Woodpecker", "Specht"], ["Camel", "Kamel"], ["Starfish", "Seestern"], ["Koala", "Koala"],
        ["Alligator", "Alligator"], ["Owl", "Eule"], ["Tiger", "Tiger"], ["Bear", "Bär"], ["BlueWhale", "Blauwal"], ["Coyote", "Kojote"],
        ["Chimpanzee", "Schimpanse"], ["Raccoon", "Waschbär"], ["Lion", "Löwe"], ["ArcticWolf", "Polarwolf"], ["Crocodile", "Krokodil"], ["Dolphin", "Delfin"],
        ["Elephant", "Elefant"], ["Squirrel", "Eichhörnchen"], ["Snake", "Schlange"], ["Kangaroo", "Känguru"], ["Hippopotamus", "Nilpferd"], ["Elk", "Elch"],
        ["Fox", "Fuchs"], ["Gorilla", "Gorilla"], ["Bat", "Fledermaus"], ["Hare", "Hase"], ["Toad", "Kröte"], ["Frog", "Frosch"], ["Deer", "Rotwild"],
        ["Rat", "Ratte"], ["Badger", "Dachs"], ["Lizard", "Eidechse"], ["Mole", "Maulwurf"], ["Hedgehog", "Igel"], ["Otter", "Fischotter"],
        ["Reindeer", "Rentier"], ["Crab", "Krabbe"], ["Fish", "Fisch"], ["Seal", "Robbe"], ["Octopus", "Oktopus"], ["Shark", "Hai"],
        ["Seahorse", "Seepferdchen"], ["Walrus", "Walross"], ["Starfish", "Seestern"], ["Whale", "Walfisch"], ["Penguin", "Pinguin"], ["Jellyfish", "Qualle"],
        ["Squid", "Tintenfisch"], ["Lobster", "Hummer"], ["Pelican", "Pelikan"], ["Clam", "Muschel"], ["Seagull", "Möwe"], ["Dolphin", "Delphin"],
        ["Shell", "Muschel"], ["SeaUrchin", "Seebarsch"], ["Cormorant", "Kormoran"], ["Otter", "Fischotter"], ["Pelican", "Pelikan"],
        ["SeaAnemone", "Seeanemone"], ["SeaTurtle", "Meeresschildkröte"], ["SeaLion", "Seelöwe"], ["Coral", "Koralle"], ["Moth", "Motte"], ["Bee", "Biene"],
        ["Butterfly", "Schmetterling"], ["Spider", "Spinne"], ["Ant", "Ameise"], ["Dragonfly", "Libelle"], ["Fly", "Fliege"], ["Mosquito", "Stechmücke"],
        ["Grasshopper", "Grashüpfer"], ["Beetle", "Käfer"], ["Cockroach", "Kakerlake"], ["Centipede", "Tausendfüßler"], ["Worm", "Wurm"], ["Louse", "Laus"]
    ]
    name = random.choice(animals)
    return name



from wiederverwendbar.task_manger.manager import Manager, ManagerSingleton
from wiederverwendbar.task_manger.task import Task
from wiederverwendbar.task_manger.trigger import Interval, EverySeconds, EveryMinutes, EveryHours, EveryDays, EveryWeeks, EveryMonths, EveryYears, At, AtNow, AtCreation



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import Optional


class Trigger(ABC):
    def __init__(self):
        self.manager = None

    @abstractmethod
    def init(self, manager):
        ...

    @abstractmethod
    def next(self, *args, **kwargs) -> datetime:
        ...


class Interval(Trigger):
    def __init__(self,
                 seconds: int = 0,
                 minutes: int = 0,
                 hours: int = 0,
                 days: int = 0,
                 weeks: int = 0,
                 months: int = 0,
                 years: int = 0):
        super().__init__()

        self.interval: Optional[int] = None
        self.seconds: int = seconds
        self.minutes: int = minutes
        self.hours: int = hours
        self.days: int = days
        self.weeks: int = weeks
        self.months: int = months
        self.years: int = years

    def init(self, manager):
        self.manager = manager

        # calculate interval in seconds
        self.interval = self.seconds
        self.interval += self.minutes * 60
        self.interval += self.hours * 60 * 60
        self.interval += self.days * 60 * 60 * 24
        self.interval += self.weeks * 60 * 60 * 24 * 7
        self.interval += self.months * 60 * 60 * 24 * 30
        self.interval += self.years * 60 * 60 * 24 * 365

    def __str__(self):
        interval = self.interval
        interval_str = "Every "
        if interval >= 60 * 60 * 24 * 365:
            years = interval // (60 * 60 * 24 * 365)
            interval_str += f"{years}y "
            interval -= years * 60 * 60 * 24 * 365
        if interval >= 60 * 60 * 24 * 30:
            months = interval // (60 * 60 * 24 * 30)
            interval_str += f"{months}M "
            interval -= months * 60 * 60 * 24 * 30
        if interval >= 60 * 60 * 24 * 7:
            weeks = interval // (60 * 60 * 24 * 7)
            interval_str += f"{weeks}w "
            interval -= weeks * 60 * 60 * 24 * 7
        if interval >= 60 * 60 * 24:
            days = interval // (60 * 60 * 24)
            interval_str += f"{days}d "
            interval -= days * 60 * 60 * 24
        if interval >= 60 * 60:
            hours = interval // (60 * 60)
            interval_str += f"{hours}h "
            interval -= hours * 60 * 60
        if interval >= 60:
            minutes = interval // 60
            interval_str += f"{minutes}m "
            interval -= minutes * 60
        if interval > 0:
            interval_str += f"{interval}s "
        interval_str = interval_str.strip()
        return interval_str

    def __int__(self):
        return self.interval

    def __add__(self, other):
        if isinstance(other, (Interval, int, float)):
            return Interval(seconds=int(self) + int(other))
        else:
            raise ValueError(f"Can't add Interval and '{type(other)}'.")

    def next(self, last: datetime) -> datetime:
        return last + timedelta(seconds=int(self))


class EverySeconds(Interval):
    def __init__(self, seconds: int):
        super().__init__(seconds=seconds)


class EveryMinutes(Interval):
    def __init__(self, minutes: int):
        super().__init__(minutes=minutes)


class EveryHours(Interval):
    def __init__(self, hours: int):
        super().__init__(hours=hours)


class EveryDays(Interval):
    def __init__(self, days: int):
        super().__init__(days=days)


class EveryWeeks(Interval):
    def __init__(self, weeks: int):
        super().__init__(weeks=weeks)


class EveryMonths(Interval):
    def __init__(self, months: int):
        super().__init__(minutes=months)


class EveryYears(Interval):
    def __init__(self, years: int):
        super().__init__(years=years)


class At(Trigger):
    def __init__(self,
                 second: Optional[int] = None,
                 minute: Optional[int] = None,
                 hour: Optional[int] = None,
                 day: Optional[int] = None,
                 month: Optional[int] = None,
                 year: Optional[int] = None,
                 delay_for_seconds: int = 0,
                 delay_for_minutes: int = 0,
                 delay_for_hours: int = 0,
                 delay_for_days: int = 0):
        super().__init__()

        self.second: Optional[int] = second
        self.minute: Optional[int] = minute
        self.hour: Optional[int] = hour
        self.day: Optional[int] = day
        self.month: Optional[int] = month
        self.year: Optional[int] = year
        self.delay_for_seconds: int = delay_for_seconds
        self.delay_for_minutes: int = delay_for_minutes
        self.delay_for_hours: int = delay_for_hours
        self.delay_for_days: int = delay_for_days

    def init(self, manager):
        self.manager = manager

        # check if some of the values are set
        if not any([self.second is not None, self.minute is not None, self.hour is not None, self.day is not None, self.month is not None, self.year is not None]):
            raise ValueError("At least one of the values must be set.")

        # validate values
        if self.second is not None and not 0 <= self.second <= 59:
            raise ValueError(f"Second must be in range 0-59. Got '{self.second}'.")
        if self.minute is not None and not 0 <= self.minute <= 59:
            raise ValueError(f"Minute must be in range 0-59. Got '{self.minute}'.")
        if self.hour is not None and not 0 <= self.hour <= 23:
            raise ValueError(f"Hour must be in range 0-23. Got '{self.hour}'.")
        if self.day is not None and not 1 <= self.day <= 31:
            raise ValueError(f"Day must be in range 1-31. Got '{self.day}'.")
        if self.month is not None and not 1 <= self.month <= 12:
            raise ValueError(f"Month must be in range 1-12. Got '{self.month}'.")
        if self.year is not None and not 1970 <= self.year <= 9999:
            raise ValueError(f"Year must be in range 1970-9999. Got '{self.year}'.")

    def __str__(self):
        at_str = "At "
        if self.year is not None:
            at_str += f"year {self.year:04} "
        if self.month is not None:
            if at_str:
                at_str += "and "
            at_str += f"month {self.month:02} "
        if self.day is not None:
            if at_str:
                at_str += "and "
            at_str += f"day {self.day:02} "
        if self.hour is not None:
            if at_str:
                at_str += "and "
            at_str += f"hour {self.hour:02} "
        if self.minute is not None:
            if at_str:
                at_str += "and "
            at_str += f"minute {self.minute:02} "
        if self.second is not None:
            if at_str:
                at_str += "and "
            at_str += f"second {self.second:02} "
        at_str = at_str.strip()
        return at_str

    def next(self) -> datetime:
        n = datetime.now().replace(microsecond=0)
        second = 0
        if self.second is not None:
            second = self.second
            second_n = n.replace(second=second)
            if second_n < n:
                n += timedelta(minutes=1)
            n = n.replace(second=second)
        minute = 0
        if self.minute is not None:
            minute = self.minute
            minute_n = n.replace(minute=minute)
            if minute_n < n:
                n += timedelta(hours=1)
            n = n.replace(minute=minute, second=second)
        hour = 0
        if self.hour is not None:
            hour = self.hour
            hour_n = n.replace(hour=hour)
            if hour_n < n:
                n += timedelta(days=1)
            n = n.replace(hour=hour, minute=minute, second=second)
        day = 1
        if self.day is not None:
            day = self.day
            day_n = n.replace(day=day)
            if day_n < n:
                while True:
                    n += timedelta(days=1)
                    if n.day == day:
                        break
            n = n.replace(day=day, hour=hour, minute=minute, second=second)
        month = 1
        if self.month is not None:
            month = self.month
            month_n = n.replace(month=month)
            if month_n < n:
                while True:
                    n += timedelta(days=1)
                    if n.month == month:
                        break
            n = n.replace(month=month, day=day, hour=hour, minute=minute, second=second)
        if self.year is not None:
            n = n.replace(year=self.year, month=month, day=day, hour=hour, minute=minute, second=second)

        return n


class AtNow(At):
    def init(self, manager):
        self.manager = manager

        now = datetime.now() + timedelta(seconds=self.delay_for_seconds,
                                         minutes=self.delay_for_minutes,
                                         hours=self.delay_for_hours,
                                         days=self.delay_for_days)
        self.second = now.second
        self.minute = now.minute
        self.hour = now.hour
        self.day = now.day
        self.month = now.month
        self.year = now.year

        super().init(manager)


class AtCreation(At):
    def init(self, manager):
        self.manager = manager

        creation_time = self.manager.creation_time + timedelta(seconds=self.delay_for_seconds,
                                                               minutes=self.delay_for_minutes,
                                                               hours=self.delay_for_hours,
                                                               days=self.delay_for_days)
        self.second = creation_time.second
        self.minute = creation_time.minute
        self.hour = creation_time.hour
        self.day = creation_time.day
        self.month = creation_time.month
        self.year = creation_time.year

        super().init(manager)



import typer
from wiederverwendbar.typer import typer_resolve_defaults

app = typer.Typer()


@app.command()
@typer_resolve_defaults
def main(a: str, name: str = typer.Argument(default="World1"), qwe: str = typer.Option(default_factory=lambda: "qwe")):
    print(f"Hello {name}")


if __name__ == "__main__":
    main(a="a1")
    app()



from typing import Type, Any, Optional, Union
from starlette_admin.fields import BaseField
from starlette_admin.contrib.mongoengine import ModelView as MongoModelView


class PropertyModelView:
    property_field_position_before: Optional[dict[str, Union[str, Any]]] = None

    def __new__(cls, *args, **kwargs):
        # override __init__ method
        __original_init__ = getattr(cls, "__init__", None)

        def __init__(self, *a, **kw):
            # convert property_field_position_before to dict[str, str]
            if self.property_field_position_before is not None:
                property_field_position_before = {}
                for attr_name, attr in self.property_field_position_before.items():
                    if isinstance(attr, str):
                        property_field_position_before[attr_name] = attr
                    else:
                        if hasattr(attr, "name"):
                            property_field_position_before[attr_name] = attr.name
                        else:
                            raise ValueError(f"{attr} must have a name attribute.")
            else:
                property_field_position_before = None
            self.property_field_position_before = property_field_position_before

            # call original __init__ method
            if __original_init__ is not None:
                __original_init__(self, *a, **kw)

            # call __post_init__ method
            if hasattr(self, "__post_init__"):
                self.__post_init__()

        cls.__init__ = __init__

        return super().__new__(cls)

    def __post_init__(self):
        if isinstance(self, MongoModelView):
            for attr_name, attr in self.document.__dict__.items():
                if not isinstance(attr, property):
                    continue
                sa_field_type: Type[BaseField] = getattr(attr.fget, "__sa_field_type__", None)
                if sa_field_type is None:
                    continue
                sa_field_kwargs: dict[str, Any] = getattr(attr.fget, "__sa_field_kwargs__")

                allow_set = True if attr.fset is not None else False

                sa_field: BaseField = sa_field_type(**sa_field_kwargs)
                if not allow_set:
                    sa_field.exclude_from_create = True
                    sa_field.exclude_from_edit = True

                # position before another field
                property_field_position_before: dict[str, str] = getattr(self, "property_field_position_before", None)
                fields: list[BaseField] = getattr(self, "fields", [])
                searchable_fields: list[str] = getattr(self, "searchable_fields", [])
                sortable_fields: list[str] = getattr(self, "sortable_fields", [])
                exclude_fields_from_create: list[str] = getattr(self, "exclude_fields_from_create", [])
                exclude_fields_from_edit: list[str] = getattr(self, "exclude_fields_from_edit", [])
                if sa_field.name in property_field_position_before:
                    position_before = property_field_position_before[sa_field.name]
                    for i, field in enumerate(fields):
                        if field.name == position_before:
                            fields.insert(i, sa_field)
                            searchable_fields.insert(i, sa_field.name)
                            sortable_fields.insert(i, sa_field.name)
                            exclude_fields_from_create.insert(i, sa_field.name)
                            exclude_fields_from_edit.insert(i, sa_field.name)
                            break
                    else:
                        raise ValueError(f"Field {position_before} not found.")
                else:
                    fields.append(sa_field)
                    searchable_fields.append(sa_field.name)
                    sortable_fields.append(sa_field.name)
                    exclude_fields_from_create.append(sa_field.name)
                    exclude_fields_from_edit.append(sa_field.name)

                self.fields = fields
                self.searchable_fields = searchable_fields
                self.sortable_fields = sortable_fields
                if not allow_set:
                    self.exclude_fields_from_create = exclude_fields_from_create
                    self.exclude_fields_from_edit = exclude_fields_from_edit
        else:
            raise NotImplementedError(f"{self.__class__.__name__} is not supported yet.")



from wiederverwendbar.pydantic.file_config import FileConfig


class Config(FileConfig):
    asd: int = 123
    qwe: str = "qwe"
    yxc: bool = False


if __name__ == '__main__':
    config = Config(file_path="test")



from enum import Enum

import uvicorn
from starlette.applications import Starlette
from starlette.responses import HTMLResponse
from starlette.routing import Route
from mongoengine import connect, Document, EmbeddedDocument, StringField, IntField, FloatField, BooleanField, ListField, DictField, EmbeddedDocumentField, \
    GenericEmbeddedDocumentField
from starlette_admin.contrib.mongoengine import ModelView

from wiederverwendbar.starlette_admin.mongoengine import Admin, GenericEmbeddedConverter

# connect to database
connect("test",
        host="localhost",
        port=27017)

# Create starlette app
app = Starlette(
    routes=[
        Route(
            "/",
            lambda r: HTMLResponse('<a href="/admin/">Click me to get to Admin!</a>'),
        ),
    ],
)

# Create admin
admin = Admin(title="Test Admin")


class Test1(EmbeddedDocument):
    meta = {"name": "test1_qwe"}

    test_1_str = StringField()
    test_1_int = IntField()
    test_1_float = FloatField()
    test_1_bool = BooleanField()


class Test2(EmbeddedDocument):
    test_2_str = StringField()
    test_2_int = IntField()
    test_2_float = FloatField()
    test_2_bool = BooleanField()
    test_2_list = ListField(StringField())
    test_2_dict = DictField()


class TestEnum(Enum):
    A = "a"
    B = "b"
    C = "c"


class Test(Document):
    meta = {"collection": "test"}

    # test_str = StringField()
    # test_int = IntField()
    # test_float = FloatField()
    # test_bool = BooleanField()
    # test_list = ListField(me.StringField())
    # test_dict = DictField()
    # test_enum = EnumField(TestEnum)
    test_gen_emb = GenericEmbeddedDocumentField(choices=[Test1, Test2], help_text="Test Generic Embedded Document Field.")
    test_gen_emb_list = ListField(GenericEmbeddedDocumentField(choices=[Test1, Test2], help_text="Test Generic Embedded Document Field."))
    # test_emb = EmbeddedDocumentField(Test2)


class TestView(ModelView):
    def __init__(self):
        super().__init__(document=Test, icon="fa fa-server", name="Test", label="Test", converter=GenericEmbeddedConverter())


# Add views to admin#
admin.add_view(TestView())

# Mount admin to app
admin.mount_to(app)

if __name__ == '__main__':
    uvicorn.run(app, host="0.0.0.0", port=8000)



import atexit
import logging
from abc import ABCMeta
from typing import Any, Optional, Union

from pydantic import BaseModel
from pydantic._internal._model_construction import ModelMetaclass

from wiederverwendbar.functions.find_class_method import find_class_method

logger = logging.getLogger(__name__)


class Singleton(ModelMetaclass):
    """
    Singleton metaclass
    """

    singleton_map: dict[str, Any] = {}
    singleton_order: dict[str, int] = {}
    delete_all_on_exit = True
    delete_ordered_on_exit = True

    def __new__(cls, name, bases, attrs, order: Optional[int] = None):
        # get __init__ method
        __init__ = attrs.pop("__init__", None)
        if __init__ is None:
            __init__ = find_class_method(bases, "__init__")

        # wrap __init__ method
        def singleton__init__(self, *args, **kwargs):
            _order = order
            if _order is None:
                _order = Singleton.get_next_order()
            self_name = self.__class__.__name__
            if _order in Singleton.singleton_order.values():
                raise RuntimeError(f"Singleton order {_order} already initialized. Use {self_name}() to get the instance.")
            if self_name not in Singleton.singleton_map:
                Singleton.singleton_map[self_name] = self
                Singleton.singleton_order[self_name] = _order
                logger.debug(f"Singleton '{self_name}' with order {_order} initialized.")
                if __init__ is not None:
                    __init__(self, *args, **kwargs)
            else:
                raise RuntimeError(f"Singleton {self_name} already initialized. Use {self_name}() to get the instance.")

        attrs["__singleton__"] = True

        # check if __singleton__ is set on bases
        singleton_set = False
        for _base in bases:
            if hasattr(_base, "__singleton__"):
                singleton_set = True
                break

        if not singleton_set:
            attrs["__init__"] = singleton__init__

        # check if BaseModel in bases
        def check_base_model(previous, _bases):
            for _b in _bases:
                if _b is previous:
                    continue
                if _b is BaseModel:
                    return True
                found = check_base_model(_b, _base.__bases__)
                if found:
                    return True
            return False

        if not check_base_model(cls, bases):
            return ABCMeta.__new__(cls, name, bases, attrs)
        else:
            return super().__new__(cls, name, bases, attrs)

    def __call__(cls, *args, init: bool = False, **kwargs):
        if init:
            if cls.__name__ not in Singleton.singleton_map:
                return super().__call__(*args, **kwargs)
            else:
                raise RuntimeError(f"Singleton {cls.__name__} already initialized. Use {cls.__name__}() to get the instance.")
        else:
            if cls.__name__ in Singleton.singleton_map:
                return cls.get_by_name(cls.__name__)
            else:
                raise RuntimeError(f"Singleton {cls.__name__} not initialized. Call {cls.__name__}(init=True) first.")

    @classmethod
    def get_all(cls, ordered: bool = True) -> dict[str, Any]:
        """
        Return all singletons in map

        :param ordered: Return ordered map
        :return: Singleton map
        """

        if ordered:
            singleton_map = {k: v for k, v in sorted(cls.singleton_map.items(), key=lambda item: cls.singleton_order[item[0]])}
        else:
            singleton_map = cls.singleton_map

        return singleton_map

    @classmethod
    def get_by_name(cls, name: str) -> Any:
        """
        Get singleton by name

        :param name: Name of singleton
        :return: SingletonInstance
        """

        current = cls.get_all().get(name, None)
        if current is None:
            raise RuntimeError(f"Singleton {name} not found.")
        return current

    @classmethod
    def get_by_type(cls, t: Union[type, str]) -> Any:
        """
        Get singleton by type

        :param t: Type of singleton
        :return: SingletonInstance
        """

        if isinstance(t, str):
            searching_name = t
        elif isinstance(t, type):
            searching_name = t.__name__
        else:
            raise TypeError(f"Type of 't' must be 'str' or 'type', not '{type(t)}'.")
        for name, instance in cls.get_all().items():
            if isinstance(t, str):
                # get all bases
                bases = instance.__class__.__bases__
                for base in bases:
                    if base.__name__ == searching_name:
                        return instance
            elif isinstance(t, type):
                if isinstance(instance, t):
                    return instance
        raise RuntimeError(f"Singleton {searching_name} not found.")

    @classmethod
    def get_by_order(cls, order: int) -> Any:
        """
        Get singleton by order

        :param order: Order of singleton
        :return: SingletonInstance
        """

        for name, _order in cls.singleton_order.items():
            if _order == order:
                return cls.get_by_name(name)
        raise RuntimeError(f"Singleton order {order} not found.")

    @classmethod
    def get_next_order(cls) -> int:
        """
        Get next order for singleton

        :return: Next order
        """

        if len(cls.singleton_order) == 0:
            return 1
        else:
            return max(cls.singleton_order.values()) + 1

    @classmethod
    def delete_all(cls, ordered: Optional[bool] = None):
        """
        Delete all singletons in map
        """

        # if ordered is None, delete by class attribute
        if ordered is None:
            ordered = cls.delete_ordered_on_exit

        singleton_names = list(cls.get_all(ordered=ordered).keys())
        singleton_names.reverse()
        for name in singleton_names:
            cls.delete_by_name(name)

    @classmethod
    def delete_by_name(cls, name: str):
        """
        Delete singleton by name

        :param name: Name of singleton
        """

        current = cls.get_all().get(name, None)
        if current is None:
            raise RuntimeError(f"Singleton {name} not found.")
        logger.debug(f"Singleton '{name}' deleted.")
        del cls.singleton_map[name]

    @classmethod
    def delete_by_type(cls, t: Union[type, str]):
        """
        Delete singleton by type

        :param t: Type of singleton
        """

        current = cls.get_by_type(t)
        if current is None:
            raise RuntimeError(f"Singleton {t} not found.")
        logger.debug(f"Singleton '{current.__class__.__name__}' deleted.")
        del cls.singleton_map[current.__class__.__name__]

    @classmethod
    def delete_by_order(cls, order: int):
        """
        Delete singleton by order

        :param order: Order of singleton
        """

        current = cls.get_by_order(order)
        if current is None:
            raise RuntimeError(f"Singleton order {order} not found.")
        logger.debug(f"Singleton '{current.__class__.__name__}' deleted.")
        del cls.singleton_map[current.__class__.__name__]


def _on_exit():
    """
    On exit hook
    """

    logger.debug("Running on exit hook.")

    if Singleton.delete_all_on_exit:
        Singleton.delete_all()


atexit.register(_on_exit)



import inspect
from functools import wraps

import typer


def typer_resolve_defaults(func):
    """
    This decorator resolves the default values of typer models.
    :param func: Function to decorate
    :return: Decorated function
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Get the function signature
        func_signature = inspect.signature(func)

        # Bind the arguments to the function signature
        bound_params = func_signature.bind(*args, **kwargs).arguments

        for param_name, param in func_signature.parameters.items():
            if param_name in bound_params:
                continue
            if param.default == param.empty:
                continue
            if type(param.default) not in [typer.models.ArgumentInfo, typer.models.OptionInfo]:
                continue
            if param.default.default_factory is not None:
                bound_params[param_name] = param.default.default_factory()
                continue
            if param.default.default is not Ellipsis:
                bound_params[param_name] = param.default.default
                continue
        return func(**bound_params)

    return wrapper



import logging

from wiederverwendbar.logger import LoggerSingleton
from wiederverwendbar.logger import LoggerSettings

import uvicorn
from fastapi import FastAPI


class TestSettings(LoggerSettings):
    qwe: str = "qwe"


test_settings = TestSettings(
    log_level="DEBUG",
    log_file=True,
    log_file_path="test.log"
)

if __name__ == '__main__':
    logger = LoggerSingleton(init=True, name="test", settings=test_settings, ignored_loggers_like=["uvicorn"])

    sub_logger = logging.getLogger("sub")

    logger.debug("debug")
    logger.info("info")
    logger.warning("warning")
    logger.error("error")
    logger.critical("critical")

    sub_logger.debug("debug")
    sub_logger.info("info")
    sub_logger.warning("warning")
    sub_logger.error("error")
    sub_logger.critical("critical")

    app = FastAPI()

    uvicorn.run(app, host="0.0.0.0", port=8000)

    print()
